require 'SketchUp.rb'
require 'Win32API'

# DLL functions
sDLLFileName = Sketchup.find_support_file "GMLSketchUpPlugin.dll", "Plugins/GMLTexturizer/"

# DLL initialization
dllInitialization = Win32API.new(sDLLFileName, "DLLInitialization", ['i','i'], 'i')

# DLL: Get texture file name
dllGetTextureFileName = Win32API.new(sDLLFileName, "DLLGetTextureFileName", [], 'p')

# DLL: Whether model should be resized
dllIsChangeModel = Win32API.new(sDLLFileName, "DLLIsChangeModel", [], 'i')

# DLL: Get texture width
dllGetTextureWidth = Win32API.new(sDLLFileName, "DLLGetTextureWidth", [], 'i')

# DLL: Get texture height
dllGetTextureHeight = Win32API.new(sDLLFileName, "DLLGetTextureHeight", [], 'i')


# Whether to check if the selected face is rectangular
m_bCheckIsRectangular = false


# Creating toolbar
toolbarGMLTexturizer = UI::Toolbar.new "GML Texturizer"

# Defining GML Texturizer command
commandGMLTexturizer = UI::Command.new("GML Texturizer") {
	# Acquiring selected face
	model = Sketchup.active_model
	selection = model.selection

	# Whether processing is available
	bDoProcess = true

	# Checking whether there was selected a single entity
	if bDoProcess && selection.length != 1
		# Displaying warning message if selection is invalid
		UI.messagebox "Invalid selection. Please select rectangular face.", MB_OK, "GML Texturizer"
		bDoProcess = false
	end

	# Checking whether the selected entity is face
	if bDoProcess
		# Selected entity
		entity = selection[0]

		if entity.typename != "Face"
			# Displaying warning message if selection is invalid
			UI.messagebox "Invalid selection. Please select rectangular face.", MB_OK, "GML Texturizer"
			bDoProcess = false
		end
	end
	
	# Checking whether the number of the selected face vertices is equal to 4
	if bDoProcess && entity.vertices.length != 4
		# Displaying warning message if selection is invalid
		UI.messagebox "Invalid selection. Please select rectangular face.", MB_OK, "GML Texturizer"
		bDoProcess = false
	end

	# Checking whether the selected face is rectangular
	if bDoProcess && m_bCheckIsRectangular
		for iIndex in 1..3
			# Current face vertice
			pointCurrent = entity.vertices[iIndex].position

			# Vector 1
			x1 = pointCurrent.x - entity.vertices[iIndex - 1].position.x
			y1 = pointCurrent.y - entity.vertices[iIndex - 1].position.y
			z1 = pointCurrent.z - entity.vertices[iIndex - 1].position.z

			# Vector 2
			if iIndex < 3
				x2 = entity.vertices[iIndex + 1].position.x - pointCurrent.x
				y2 = entity.vertices[iIndex + 1].position.y - pointCurrent.y
				z2 = entity.vertices[iIndex + 1].position.z - pointCurrent.z
			else
				x2 = entity.vertices[0].position.x - pointCurrent.x
				y2 = entity.vertices[0].position.y - pointCurrent.y
				z2 = entity.vertices[0].position.z - pointCurrent.z
			end

			# Angle cosine
			dAngleCos = (x1 * x2 + y1 * y2 + z1 * z2) / (Math.sqrt(x1 * x1 + y1 * y1 + z1 * z1) * Math.sqrt(x2 * x2 + y2 * y2 + z2 * z2))

			# Angle cosine absolute value
			if dAngleCos < 0
				dAngleCos = -dAngleCos
			end

			# Threshold for angle cosine
			dAngleCosThreshold = 0.01

			# Checking whether angle cosine is less than threshold
			if dAngleCos > dAngleCosThreshold
				# Displaying warning message if selection is invalid
				UI.messagebox "Invalid selection. Please select rectangular face.", MB_OK, "GML Texturizer"
				bDoProcess = false
				break
			end
		end
	end

	if bDoProcess
		# Texture coordinates
		texcoord = []
		texcoord[1] = [1,1]
		texcoord[3] = [0,1]
		texcoord[5] = [0,0]
		texcoord[7] = [1,0]

		# Face vertices
		point0 = entity.vertices[0].position
		point1 = entity.vertices[1].position
		point2 = entity.vertices[2].position
		point3 = entity.vertices[3].position

		# Top edge index
		iTopEdgeIndex = 0
		# Side edges indices
		iSideEdgeIndex = []

		# Estimating proper vertices order for vertical clockwise rendering
		if point0.z >= point1.z && point0.z >= point2.z && point0.z >= point3.z
			if point1.z >= point2.z && point1.z >= point3.z
				iTopEdgeIndex = 0
				iSideEdgeIndex[0] = 3
				iSideEdgeIndex[1] = 1
				texcoord[0] = entity.vertices[0].position
				texcoord[2] = entity.vertices[1].position
				texcoord[4] = entity.vertices[2].position
				texcoord[6] = entity.vertices[3].position
			else
				iTopEdgeIndex = 3
				iSideEdgeIndex[0] = 2
				iSideEdgeIndex[1] = 0
				texcoord[0] = entity.vertices[3].position
				texcoord[2] = entity.vertices[0].position
				texcoord[4] = entity.vertices[1].position
				texcoord[6] = entity.vertices[2].position
			end
		else
			if point1.z >= point0.z && point1.z >= point3.z
				iTopEdgeIndex = 1
				iSideEdgeIndex[0] = 0
				iSideEdgeIndex[1] = 2
				texcoord[0] = entity.vertices[1].position
				texcoord[2] = entity.vertices[2].position
				texcoord[4] = entity.vertices[3].position
				texcoord[6] = entity.vertices[0].position
			else
				iTopEdgeIndex = 2
				iSideEdgeIndex[0] = 1
				iSideEdgeIndex[1] = 3
				texcoord[0] = entity.vertices[2].position
				texcoord[2] = entity.vertices[3].position
				texcoord[4] = entity.vertices[0].position
				texcoord[6] = entity.vertices[1].position
			end
		end

		# Estimating texture size	
		iTextureWidth = Math.sqrt((texcoord[0].x - texcoord[2].x) * (texcoord[0].x - texcoord[2].x) + (texcoord[0].y - texcoord[2].y) * (texcoord[0].y - texcoord[2].y) + (texcoord[0].z - texcoord[2].z) * (texcoord[0].z - texcoord[2].z))
		iTextureHeight = Math.sqrt((texcoord[0].x - texcoord[6].x) * (texcoord[0].x - texcoord[6].x) + (texcoord[0].y - texcoord[6].y) * (texcoord[0].y - texcoord[6].y) + (texcoord[0].z - texcoord[6].z) * (texcoord[0].z - texcoord[6].z))

		# Open GML Texturizer main dialog
		iResult = dllInitialization.Call(iTextureWidth, iTextureHeight)
		if iResult == 1
			# Adding new material to the model
			materials = model.materials
			materialNew = materials.add 'Texture1'
			materialNew.texture = dllGetTextureFileName.Call()

			# Whether model should be resized
			iResizeModel = dllIsChangeModel.Call()
			if iResizeModel == 1 # Change width
				# Get texture size
				iNewTextureWidth = dllGetTextureWidth.Call()
				iNewTextureHeight = dllGetTextureHeight.Call()

				# Get faces common to the first side edge
				facesCommon = entity.edges[iSideEdgeIndex[0]].faces

				if facesCommon.length > 1
					# Choosing side face
					if entity == facesCommon[0]
						faceSide = facesCommon[1]
					else
						faceSide = facesCommon[0]
					end

					# If common face does not have texture
					if !faceSide.material
						iProperSideIndex = 0
					else
						# Get faces common to the second side edge
						facesCommon2 = entity.edges[iSideEdgeIndex[1]].faces

						if facesCommon2.length > 1
							# Choosing side face
							if entity == facesCommon2[0]
								faceSide = facesCommon2[1]
							else
								faceSide = facesCommon2[0]
							end

							iProperSideIndex = 1
						else
							iProperSideIndex = 0
						end
					end
				else
					# Get faces common to the second side edge
					facesCommon2 = entity.edges[iSideEdgeIndex[1]].faces

					if facesCommon2.length > 1
						# Choosing side face
						if entity == facesCommon2[0]
							faceSide = facesCommon2[1]
						else
							faceSide = facesCommon2[0]
						end

						iProperSideIndex = 1
					else
						iProperSideIndex = -1
					end
				end

				if iProperSideIndex >= 0
					# Using push/pull tool to adjust model size
					iDistance = iTextureHeight * iNewTextureWidth / iNewTextureHeight - iTextureWidth
					faceSide.pushpull iDistance, false

					# Acquiring selected face new coordinates for upper vertices
					if iTopEdgeIndex == 0
						texcoord[0] = entity.vertices[0].position
						texcoord[2] = entity.vertices[1].position
						texcoord[4] = entity.vertices[2].position
						texcoord[6] = entity.vertices[3].position
					elsif iTopEdgeIndex == 1
						texcoord[0] = entity.vertices[1].position
						texcoord[2] = entity.vertices[2].position
						texcoord[4] = entity.vertices[3].position
						texcoord[6] = entity.vertices[0].position
					elsif iTopEdgeIndex == 2
						texcoord[0] = entity.vertices[2].position
						texcoord[2] = entity.vertices[3].position
						texcoord[4] = entity.vertices[0].position
						texcoord[6] = entity.vertices[1].position
					else
						texcoord[0] = entity.vertices[3].position
						texcoord[2] = entity.vertices[0].position
						texcoord[4] = entity.vertices[1].position
						texcoord[6] = entity.vertices[2].position
					end
				end
			elsif iResizeModel == 2 # Change height
				# Get texture size
				iNewTextureWidth = dllGetTextureWidth.Call()
				iNewTextureHeight = dllGetTextureHeight.Call()

				# Get faces common to the top edge
				facesCommon = entity.edges[iTopEdgeIndex].faces

				if facesCommon.length > 1
					# Choosing top face
					if entity == facesCommon[0]
						faceTop = facesCommon[1]
					else
						faceTop = facesCommon[0]
					end

					# Using push/pull tool to adjust model size
					iDistance = iTextureWidth * iNewTextureHeight / iNewTextureWidth - iTextureHeight
					faceTop.pushpull iDistance, false

					# Acquiring selected face new coordinates for upper vertices
					if iTopEdgeIndex == 0
						texcoord[0] = entity.vertices[0].position
						texcoord[2] = entity.vertices[1].position
					elsif iTopEdgeIndex == 1
						texcoord[0] = entity.vertices[1].position
						texcoord[2] = entity.vertices[2].position
					elsif iTopEdgeIndex == 2
						texcoord[0] = entity.vertices[2].position
						texcoord[2] = entity.vertices[3].position
					else
						texcoord[0] = entity.vertices[3].position
						texcoord[2] = entity.vertices[0].position
					end
				end
			end

			# Placing texture on the face
			entity.position_material materialNew, texcoord, true
		end
	end

	# A copy of GML Texturizer is not running
	bIsRunning = false
	}

# Command button appearance
sGMLTexturizerBMP = Sketchup.find_support_file "GMLTexturizer.png", "Plugins/GMLTexturizer/"
commandGMLTexturizer.small_icon = sGMLTexturizerBMP
commandGMLTexturizer.large_icon = sGMLTexturizerBMP
commandGMLTexturizer.tooltip = "GML Texturizer"
commandGMLTexturizer.status_bar_text = "Create rectangle texture of proper size from image"
commandGMLTexturizer.menu_text = "GML Texturizer"

# Adding command button to the toolbar
toolbarGMLTexturizer.add_item commandGMLTexturizer


# Defining GML Texture Fit command
commandGMLTextureFit = UI::Command.new("GML Texture Fit") {
	# Whether processing is available
	bDoProcess = true

	# Acquiring selected face
	model = Sketchup.active_model
	selection = model.selection
	# Checking whether there was selected a single entity
	if selection.length != 1
		# Displaying warning message if selection is invalid
		UI.messagebox "Invalid selection. Please select rectangular face with texture.", MB_OK, "GML Texture Fit"
		bDoProcess = false
	end

	# Checking whether the selected entity is face
	if bDoProcess
		# Selected entity
		entity = selection[0]

		if entity.typename != "Face"
			# Displaying warning message if selection is invalid
			UI.messagebox "Invalid selection. Please select rectangular face with texture.", MB_OK, "GML Texture Fit"
			bDoProcess = false
		end
	end
	
	# Checking whether the number of the selected face vertices is equal to 4
	if bDoProcess && entity.vertices.length != 4
		# Displaying warning message if selection is invalid
		UI.messagebox "Invalid selection. Please select rectangular face with texture.", MB_OK, "GML Texture Fit"
		bDoProcess = false
	end

	# Checking whether the selected face is rectangular
	if bDoProcess && m_bCheckIsRectangular
		for iIndex in 1..3
			# Current face vertice
			pointCurrent = entity.vertices[iIndex].position

			# Vector 1
			x1 = pointCurrent.x - entity.vertices[iIndex - 1].position.x
			y1 = pointCurrent.y - entity.vertices[iIndex - 1].position.y
			z1 = pointCurrent.z - entity.vertices[iIndex - 1].position.z

			# Vector 2
			if iIndex < 3
				x2 = entity.vertices[iIndex + 1].position.x - pointCurrent.x
				y2 = entity.vertices[iIndex + 1].position.y - pointCurrent.y
				z2 = entity.vertices[iIndex + 1].position.z - pointCurrent.z
			else
				x2 = entity.vertices[0].position.x - pointCurrent.x
				y2 = entity.vertices[0].position.y - pointCurrent.y
				z2 = entity.vertices[0].position.z - pointCurrent.z
			end

			# Angle cosine
			dAngleCos = (x1 * x2 + y1 * y2 + z1 * z2) / (Math.sqrt(x1 * x1 + y1 * y1 + z1 * z1) * Math.sqrt(x2 * x2 + y2 * y2 + z2 * z2))

			# Angle cosine absolute value
			if dAngleCos < 0
				dAngleCos = -dAngleCos
			end

			# Threshold for angle cosine
			dAngleCosThreshold = 0.01

			# Checking whether angle cosine is less than threshold
			if dAngleCos > dAngleCosThreshold
				# Displaying warning message if selection is invalid
				UI.messagebox "Invalid selection. Please select rectangular face with texture.", MB_OK, "GML Texture Fit"
				bDoProcess = false
				break
			end
		end
	end

	if bDoProcess && !entity.material
		# Displaying warning message if selection is invalid
		UI.messagebox "Invalid selection. Please select rectangular face with texture.", MB_OK, "GML Texture Fit"
		bDoProcess = false
	end

	if bDoProcess
		# Texture coordinates
		texcoord = []
		texcoord[1] = [1,1]
		texcoord[3] = [0,1]
		texcoord[5] = [0,0]
		texcoord[7] = [1,0]

		# Face vertices
		point0 = entity.vertices[0].position
		point1 = entity.vertices[1].position
		point2 = entity.vertices[2].position
		point3 = entity.vertices[3].position

		# Top edge index
		iTopEdgeIndex = 0

		# Estimating proper vertices order for vertical clockwise rendering
		if point0.z >= point1.z && point0.z >= point2.z && point0.z >= point3.z
			if point1.z >= point2.z && point1.z >= point3.z
				iTopEdgeIndex = 0
				texcoord[0] = entity.vertices[0].position
				texcoord[2] = entity.vertices[1].position
				texcoord[4] = entity.vertices[2].position
				texcoord[6] = entity.vertices[3].position
			else
				iTopEdgeIndex = 3
				texcoord[0] = entity.vertices[3].position
				texcoord[2] = entity.vertices[0].position
				texcoord[4] = entity.vertices[1].position
				texcoord[6] = entity.vertices[2].position
			end
		else
			if point1.z >= point0.z && point1.z >= point3.z
				iTopEdgeIndex = 1
				texcoord[0] = entity.vertices[1].position
				texcoord[2] = entity.vertices[2].position
				texcoord[4] = entity.vertices[3].position
				texcoord[6] = entity.vertices[0].position
			else
				iTopEdgeIndex = 2
				texcoord[0] = entity.vertices[2].position
				texcoord[2] = entity.vertices[3].position
				texcoord[4] = entity.vertices[0].position
				texcoord[6] = entity.vertices[1].position
			end
		end

		# Placing texture on the face
		entity.position_material entity.material, texcoord, true
	end
	}

# Command button appearance
sGMLTextureFitBMP = Sketchup.find_support_file "GMLTextureFit.png", "Plugins/GMLTexturizer/"
commandGMLTextureFit.small_icon = sGMLTextureFitBMP
commandGMLTextureFit.large_icon = sGMLTextureFitBMP
commandGMLTextureFit.tooltip = "GML Texture Fit"
commandGMLTextureFit.status_bar_text = "Properly fit current texture to the selected face"
commandGMLTextureFit.menu_text = "GML Texture Fit"

# Adding command button to the toolbar
toolbarGMLTexturizer.add_item commandGMLTextureFit

# Showing toolbar
toolbarGMLTexturizer.show

file_loaded("GMLTexturizer.rb")